(
 (c-mode . (
            (c-file-style . "K&R")
            (indent-tabs-mode . nil)
            (c-indent-level . 4)
            (c-basic-offset . 4)
            ))
 (html-mode . (
	       (indent-tabs-mode . nil)
	       ))
 (sh-mode . (
	     (indent-tabs-mode . nil)
	     ))
 (nxml-mode . (
	       (indent-tabs-mode . nil)
	       ))
 (perl-mode . (
	       (indent-tabs-mode . nil)
	       ))
 )
